unit oTermMappingTableEntry;

interface

uses
  Classes,
  iCPTerminology,
  ComCtrls,
  XMLIntf,
  uXML,
  SysUtils;

type
  TTermMappingTableEntry = class(TInterfacedObject, ITermMappingTableEntry)
  private
    fTerminology: ITerminologyController;
    fTermMappingTable: ITermMappingTable;
    fXMLNode: IXMLNode;
    fChangeMade: boolean;

    fVendorKey: string;
    fDoNotProcess: boolean;
    fMappedTerm: array[TTermType] of ITerm;

    fObservation: ITerm;
    fUnit: ITerm;
    fMethod: ITerm;
    fLocation: ITerm;
    fQuality: ITerm;
    fProduct: ITerm;

    function getVendorKey: string; stdcall;
    function getDoNotProcess: boolean; stdcall;

    function getMappedTerm(aTermType: TTermType): ITerm; stdcall;
    function getMappedTermID(aTermType: TTermType): string; stdcall;

    function getXMLNode: IXMLNode; stdcall;
    function getMappingIDs: string; stdcall;
    function getMappingDisplay: string; stdcall;

    procedure setVendorKey(const aValue: string); stdcall;
    procedure setDoNotProcess(const aValue: boolean); stdcall;
    procedure setMappedTerm(const aTermType: TTermType; const aValue: ITerm); stdcall;

    procedure Load; stdcall;
    procedure Save; stdcall;

    function AddToList(aList: TStrings): integer; overload; stdcall;
    function AddToList(aListView: TListView; aImageIndex: integer = -1): TListItem; overload; stdcall;
  public
    constructor Create(aTermMappingTable: ITermMappingTable; aTerminology: ITerminologyController; aXMLNode: IXMLNode);
    destructor Destroy; override;
  end;

implementation

const
  XML_VENDOR_KEY = 'VENDOR_KEY';
  XML_DO_NOT_PROCESS = 'DO_NOT_PROCESS';

{ TTermMappingTableEntry }

constructor TTermMappingTableEntry.Create(aTermMappingTable: ITermMappingTable; aTerminology: ITerminologyController; aXMLNode: IXMLNode);
var
  aTermType: TTermType;
begin
  inherited Create;
  fTerminology := aTerminology;
  fTermMappingTable := aTermMappingTable;
  fXMLNode := aXMLNode;

  if fXMLNode.ChildNodes.FindNode(XML_VENDOR_KEY) = nil then
    fXMLNode.ChildValues[XML_VENDOR_KEY] := 'New Vendor Key';

  if fXMLNode.ChildNodes.FindNode(XML_DO_NOT_PROCESS) = nil then
    fXMLNode.ChildValues[XML_DO_NOT_PROCESS] := 1;

  fVendorKey := fXMLNode.ChildValues[XML_VENDOR_KEY];
  fDoNotProcess := XMLValueAsInt(fXMLNode, XML_DO_NOT_PROCESS, 1) = 1;

  for aTermType := Low(TTermType) to High(TTermType) do
    fMappedTerm[aTermType] := nil;

  fObservation := nil;
  fUnit := nil;
  fMethod := nil;
  fLocation := nil;
  fQuality := nil;
  fProduct := nil;
end;

destructor TTermMappingTableEntry.Destroy;
begin
  fTerminology := nil;
  inherited;
end;

function TTermMappingTableEntry.AddToList(aList: TStrings): integer;
begin
  Result := aList.AddObject(fVendorKey, Self);
end;

function TTermMappingTableEntry.AddToList(aListView: TListView; aImageIndex: integer = -1): TListItem;
begin
  Result := aListView.Items.Add;
  Result.Caption := fVendorKey;
  Result.Data := Self;
  Result.ImageIndex := aImageIndex;
end;

procedure TTermMappingTableEntry.Load;
var
  aTermType: TTermType;
begin
  try
    fVendorKey := XMLValueAsStr(fXMLNode, XML_VENDOR_KEY);
    fDoNotProcess := XMLValueAsInt(fXMLNode, XML_DO_NOT_PROCESS) = 1;

    {This converts old style data to new style}
    if XMLValueAsStr(fXMLNode, 'OBSERVATION_ID') <> '' then
      begin
        fXMLNode.ChildValues[TERM_TYPE_XML[ttObservation]] := XMLValueAsStr(fXMLNode, 'OBSERVATION_ID');
        fXMLNode.ChildValues['OBSERVATION_ID'] := '';
      end;

    for aTermType := Low(TTermType) to High(TTermType) do
      fMappedTerm[aTermType] := fTerminology.getTerm(XMLValueAsStr(fXMLNode, TERM_TYPE_XML[aTermType]));

    fChangeMade := False;
  except
    raise ETerminologyLoadError.Create('Error loading mapping table entry');
  end;
end;

procedure TTermMappingTableEntry.Save;
var
  aTermType: TTermType;
begin
  fXMLNode.ChildNodes.Clear;
  fXMLNode.ChildValues[XML_VENDOR_KEY] := fVendorKey;
  fXMLNode.ChildValues[XML_DO_NOT_PROCESS] := BOOLEAN_TO_INT[fDoNotProcess];

  for aTermType := Low(TTermType) to High(TTermType) do
    if assigned(fMappedTerm[aTermType]) then
      fXMLNode.ChildValues[TERM_TYPE_XML[aTermType]] := fMappedTerm[aTermType].getID;

  fChangeMade := False;
end;

function TTermMappingTableEntry.getMappingDisplay: string;
var
  aTermType: TTermType;
begin
  Result := '';
  if fDoNotProcess then
    Result := 'Key does not process'
  else
    for aTermType := ttObservation to High(TTermType) do
      if Assigned(getMappedTerm(aTermType)) then
        if Result <> '' then
          Result := Format('%s, %s: %s', [Result, Copy(TERM_TYPE[aTermType], 1, 1), getMappedTerm(aTermType).getTerm])
        else
          Result := Format('%s: %s', [Copy(TERM_TYPE[aTermType], 1, 1), getMappedTerm(aTermType).getTerm]);
end;

function TTermMappingTableEntry.getMappingIDs: string;
begin
  Result := 'fix getMappingIDs method';
end;

function TTermMappingTableEntry.getXMLNode: IXMLNode;
begin
  Result := fXMLNode;
end;

{ Getters  and Setters }

function TTermMappingTableEntry.getMappedTerm(aTermType: TTermType): ITerm;
begin
  Result := fMappedTerm[aTermType];
end;

function TTermMappingTableEntry.getMappedTermID(aTermType: TTermType): string;
begin
  if Assigned(fMappedTerm[aTermType]) then
    Result := fMappedTerm[aTermType].getID
  else
    Result := '';
end;

function TTermMappingTableEntry.getDoNotProcess: boolean;
begin
  Result := fDoNotProcess;
end;

function TTermMappingTableEntry.getVendorKey: string;
begin
  Result := fVendorKey;
end;

procedure TTermMappingTableEntry.setDoNotProcess(const aValue: boolean);
begin
  fDoNotProcess := aValue;
  fTermMappingTable.setChangeMade(True);
end;

procedure TTermMappingTableEntry.setMappedTerm(const aTermType: TTermType; const aValue: ITerm);
begin
  fMappedTerm[aTermType] := aValue;
end;

procedure TTermMappingTableEntry.setVendorKey(const aValue: string);
begin
  fVendorKey := aValue;
  fTermMappingTable.setChangeMade(True);
end;

end.

