unit uDevices;

interface

uses
  Classes, SysUtils, IniFiles, uPatients, ExtCtrls;

type
  TObservation = class;
  TObservationEvent = procedure(Sender: TObservation) of object;

  TValueType = (
    vtAddress,
    vtCodedEntry,
    vtCodedElementWFormattedValues,
    vtCompositeIDWCheckDigit,
    vtCompositeIDAndName,
    vtCompositePrice,
    vtExtendedCompositeIDWCheckDigit,
    vtDate,
    vtEncapsulatedData,
    vtFormattedTextDisplay,
    vtMoney,
    vtNumeric,
    vtPersonName,
    vtReferencePointer,
    vtStructuredNumeric,
    vtStringData,
    vtTime,
    vtTelephoneNumber,
    vtTimeStamp,
    vtTextDataDisplay,
    vtExtendedAddress,
    vtExtendedCompositeNameNumberForPersons,
    vtExtendedCompositeNameNumberForOrganizations,
    vtExtendedPersonName,
    vtExtendedTelecommunicationsNumber
  );

  TAbnormalType = (
    atLow,
    atHigh,
    atVeryLow,
    atVeryHigh,
    atLessThan,
    atGreaterThan,
    atNormal,
    atAbnormal,
    atVeryAbnormal,
    atNull,
    atSignificantUp,
    atSignificantDown,
    atBetter,
    atWorse,
    atSuceptible,
    atResistant,
    atIntermediate,
    atModeratelySucceptible,
    atVerySucceptible
  );

  TAbnormalTypeSet = set of TAbnormalType;

  TObservation = class(TCollectionItem)
  private
    FFrequency: Integer;
    FType: TValueType;
    FDescription: string;
    FPrecision: Integer;
    FErrorData: string;
    FErrorPercentage: Integer;
    FRangeHi: Single;
    FRangeLo: Single;
    FObservationIdentifier: string;
    FObservationSubID: string;
    FObservationValue: string;
    FAbnormalFlags: TAbnormalTypeSet;
    FProbability: string;
    FAbnormalNature: string;
    FObservationResultStatus: string;
    FRefRangeEffDate: TDateTime;
    FUserDefinedAccessChecks: string;
    FProducerID: string;
    FResponsibleObserver: string;
    FObservationMethod: string;
    FEquipmentInstanceIdentifier: string;
    //
    {
      If FDirectValue is not null, it will completely override all of the
      other values entered above.  It will just be repeated back
      verbatim when the time comes to generate a message.
    }
    FDirectValue: string;
    function GetSegment: string;
  protected
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
  published
    property ValueType: TValueType read FType write FType;
    property Description: string read FDescription write FDescription;
    property RangeHi: Single read FRangeHi write FRangeHi;
    property RangeLo: Single read FRangeLo write FRangeLo;
    property ObservationIdentifier: string read FObservationIdentifier write FObservationIdentifier;
    property ObservationSubID: string read FObservationSubID write FObservationSubID;
    property ObservationValue: string read FObservationValue write FObservationValue;
    property AbnormalFlags: TAbnormalTypeSet read FAbnormalFlags write FAbnormalFlags;
    property Probability: string read FProbability write FProbability;
    property AbnormalNature: string read FAbnormalNature write FAbnormalNature;
    property ObservationResultStatus: string read FObservationResultStatus write FObservationResultStatus;
    property RefRangeEffDate: TDateTime read FRefRangeEffDate write FRefRangeEffDate;
    property UserDefinedAccessChecks: string read FUserDefinedAccessChecks write FUserDefinedAccessChecks;
    property ProducerID: string read FProducerID write FProducerID;
    property ResponsibleObserver: string read FResponsibleObserver write FResponsibleObserver;
    property ObservationMethod: string read FObservationmethod write FObservationMethod;
    property EquipmentInstanceIdentifier: string read FEquipmentInstanceIdentifier
      write FEquipmentInstanceIdentifier;
    property DirectValue: string read FDirectValue write FDirectValue;
    property Frequency: Integer read FFrequency write FFrequency;
    property Segment: string read GetSegment;
  end;

  TDevice = class;
  TObservationList = class(TCollection)
  private
    FDelimiters: string;
    FParent: TDevice;
  protected
    function GetItem(Index: Integer): TObservation;
    procedure SetItem(Index: Integer; Value: TObservation);
  public
    destructor Destroy; override;
    function Add: TObservation;
    function Insert(Index: Integer): TObservation;
    property Parent: TDevice read FParent write FParent;
    property Items[Index: Integer]: TObservation read GetItem write SetItem; default;
  published
  end;

  TDevice = class(TCollectionItem)
  private
    FInterval: Integer;
    FFieldSeperator: char;
    FEncodingChars: string;
    FSendingApplication: string;
    FSendingFacility: string;
    FReceivingApplication: string;
    FReceivingFacility: string;
    FMessageType: string;
    FEventType: string;
    FMessageControlID: string;
    FProcessingCode: string;
    FVersionID: string;
    FSendAllObservationsAtOnce: boolean;

    FDescription: string;
    FOBRSegment: string;
    FORCSegment: string;
    FObservationList: TObservationList;
  protected
    function GetHeader: string;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
  published
    // This is the human readable description.
    property Description: string read FDescription write FDescription;

    // This is the actual HL7 description.
    property FieldSeperator: char read FFieldSeperator write FFieldSeperator;
    property EncodingChars: string read FEncodingChars write FEncodingChars;
    property SendingApplication: string read FSendingApplication write FSendingApplication;
    property SendingFacility: string read FSendingFacility write FSendingFacility;
    property ReceivingApplication: string read FReceivingApplication write FReceivingApplication;
    property ReceivingFacility: string read FReceivingFacility write FReceivingFacility;
    property MessageType: string read FMessageType write FMessageType;
    property EventType: string read FEventType write FEventType;
    property MessageControlID: string read FMessageControlID write FMessageControlID;
    property ProcessingCode: string read FProcessingCode write FProcessingCode;
    property VersionID: string read FVersionID write FVersionID;
    property SendAllObservationsAtOnce: boolean read FSendAllObservationsAtOnce write FSendAllObservationsAtOnce;
    property Interval: integer read FInterval write FInterval;

    property Header: string read GetHeader;
    property OBRSegment: string read FOBRSegment write FOBRSegment;
    property ORCSegment: string read FORCSegment write FORCSegment;
    property Observations: TObservationList read FObservationList write FObservationList;
  end;

  TDeviceList = class(TCollection)
  private
    FDelimiters: string;
    FHashList: THashedStringList;
  protected
    function GetItem(Index: Integer): TDevice;
    procedure SetItem(Index: Integer; Value: TDevice);
    procedure SetDelimiters(const Value: string);
  public
    constructor Create(ItemClass: TCollectionItemClass);
    destructor Destroy; override;
    function Add: TDevice; overload;
    function Add(const Name: string): TDevice; overload;
    function Insert(Index: Integer): TDevice;
    procedure Load(AFilename: string);
    procedure Save(AFilename: string);
    function IndexOfName(AName: string): integer;
    function Find(AName: string): TDevice;
    property Items[Index: Integer]: TDevice read GetItem write SetItem; default;
  published
    property Delimiters: string read FDelimiters write SetDelimiters;
  end;

  TUnownedDeviceList = class(TList)
  private
  protected
    function Get(Index: Integer): TDevice;
    procedure Put(Index: Integer; Device: TDevice);
  public
    function Add(Device: TDevice): Integer;
    function Extract(Device: TDevice): TDevice;
    function First: TDevice;
    function IndexOf(Device: TDevice): Integer;
    procedure Insert(Index: Integer; Device: TDevice);
    function Last: TDevice;
    function Remove(Device: TDevice): Integer;
    property Items[Index: Integer]: TDevice read Get write Put; default;
  published
  end;

  EDeviceException = class(Exception);

  EDeviceNameException = class(EDeviceException);

implementation

uses
  XMLDoc, XMLIntf, XMLDom, Variants, Math, uConstants;


function ArrayIndexOf(const AConstArray: array of String; const AValue: string): Integer;
var
  I : Integer;
begin
  for I := Low(AConstArray) to High(AConstArray) do
  begin
    if AConstArray[I] = AValue then
    begin
      result := I;
      exit;
    end;
  end;
  result := -1;
end;

function StringToAbnormalTypes(const Value: string):TAbnormalTypeSet;
  function intStrToAbnormalTypes(const Value: string): TAbnormalTypeSet;
  var
    AbTypeStr: string;
    AbTypeIndex: Integer;
    AbType: TAbnormalType;
  begin
    // Takes a space delimited list of abnormal type codes and converts them
    // to a set of TAbnormalTypes
    if Length(Value) = 0 then
    begin
      result := [];
      exit;
    end;

    AbTypeStr := copy(Value, 1, Pos(' ', Value));
    AbTypeIndex := ArrayIndexOf(ABNORMAL_TYPE_ID, AbTypeStr);
    if AbTypeIndex > -1 then
    begin
      AbType := TAbnormalType(AbTypeIndex);
      result := result + [AbType];
    end;

    result := result + intStrToAbnormalTypes(copy(Value, pos(' ', Value), Length(Value)));
  end;
begin
  result := intStrToAbnormalTypes(Value + ' ');
end;

function AbnormalTypesToString(const Value: TAbnormalTypeSet): string;
begin
  if atLow in Value then
    result := result + ABNORMAL_TYPE_ID[0] + ' ';
  if atHigh in Value then
    result := result + ABNORMAL_TYPE_ID[1] + ' ';
  if atVeryLow in Value then
    result := result + ABNORMAL_TYPE_ID[2] + ' ';
  if atVeryHigh in Value then
    result := result + ABNORMAL_TYPE_ID[3] + ' ';
  if atLessThan in Value then
    result := result + ABNORMAL_TYPE_ID[4] + ' ';
  if atGreaterThan in Value then
    result := result + ABNORMAL_TYPE_ID[5] + ' ';
  if atNormal in Value then
    result := result + ABNORMAL_TYPE_ID[6] + ' ';
  if atAbnormal in Value then
    result := result + ABNORMAL_TYPE_ID[7] + ' ';
  if atVeryAbnormal in Value then
    result := result + ABNORMAL_TYPE_ID[8] + ' ';
  if atNull in Value then
    result := result + ABNORMAL_TYPE_ID[9] + ' ';
  if atSignificantUp in Value then
    result := result + ABNORMAL_TYPE_ID[10] + ' ';
  if atSignificantDown in Value then
    result := result + ABNORMAL_TYPE_ID[11] + ' ';
  if atBetter in Value then
    result := result + ABNORMAL_TYPE_ID[12] + ' ';
  if atWorse in Value then
    result := result + ABNORMAL_TYPE_ID[13] + ' ';
  if atSuceptible in Value then
    result := result + ABNORMAL_TYPE_ID[14] + ' ';
  if atResistant in Value then
    result := result + ABNORMAL_TYPE_ID[15] + ' ';
  if atIntermediate in Value then
    result := result + ABNORMAL_TYPE_ID[16] + ' ';
  if atModeratelySucceptible in Value then
    result := result + ABNORMAL_TYPE_ID[17] + ' ';
  if atVerySucceptible in Value then
    result := result + ABNORMAL_TYPE_ID[18] + ' ';
end;

{ TDeviceList }

function TDeviceList.Add: TDevice;
begin
  raise EDeviceNameException.Create('Do not use TDeviceList.Add!');
end;

function TDeviceList.Add(const Name: string): TDevice;
begin
  if Name = '' then
    raise EDeviceNameException.Create('The device name must not be blank.');

  if IndexOfName(Name) = -1 then
  begin
    result := TDevice(inherited Add);
    result.Description := Name;
    FHashList.AddObject(result.Description + '=', result);
  end
  else
    raise EDeviceNameException.Create('The device name must be unique among all devices.');
end;

constructor TDeviceList.Create(ItemClass: TCollectionItemClass);
begin
  inherited Create(ItemClass);
  FHashList := THashedStringList.Create;
end;

destructor TDeviceList.Destroy;
begin
  FreeAndNil(FHashList);
  inherited;
end;

function TDeviceList.Find(AName: string): TDevice;
begin
  result := TDevice(FHashList.Objects[FHashList.IndexOfName(AName)]);
end;

function TDeviceList.GetItem(Index: Integer): TDevice;
begin
  result := TDevice(inherited GetItem(Index));
end;

function TDeviceList.IndexOfName(AName: string): integer;
begin
  result := FHashList.IndexOfName(AName);
end;

function TDeviceList.Insert(Index: Integer): TDevice;
begin
  result := TDevice(inherited Insert(Index));
end;

procedure TDeviceList.Load(AFilename: string);
var
  XMLDoc: IXMLDocument;
  I, J: Integer;
  RootNode: IXMLNode;
  DeviceNode: IXMLNode;
  DataNode: IXMLNode;
  DTypesNode: IXMLNode;
  DTypeNode: IXMLNode;
  Device: TDevice;
  Observation: TObservation;
  Pipe: string;
begin
  Clear;

  XMLDoc := TXMLDocument.Create(nil);
  try
    DefaultDOMVendor := 'MSXML';

    XMLDoc.LoadFromFile(AFilename);
    XMLDoc.Options := XMLDoc.Options + [doNodeAutoIndent];
    XMLDoc.Active := true;
    XMLDoc.Encoding := 'UTF-8';

    RootNode := XMLDoc.DocumentElement;
    if RootNode.NodeName <> DEVICES then
      raise EDeviceException.Create('Invalid device file');

    for I := 0 to RootNode.ChildNodes.Count - 1 do
    begin
      DeviceNode := RootNode.ChildNodes[I];

      DataNode := DeviceNode.ChildNodes.FindNode(NAME);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device := Add(DataNode.NodeValue)
      else
      begin
        raise EDeviceException.Create('Unassigned data node in TDeviceList.Load.');
        //Since the compiler isn't smart enough to realize that raising an exception here
        // will pop us completely out of the load procedure, we have to nil out the
        // device object below to get rid of the unassigned object warning.
        Device := nil;
      end;

      DataNode := DeviceNode.ChildNodes.FindNode(OBRSEGMENT);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FOBRSegment := DataNode.NodeValue;




      DataNode := DeviceNode.ChildNodes.FindNode(ORCSEGMENT);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FORCSegment := DataNode.NodeValue;
        
      {
      FFieldSeperator: char;
      FEncodingChars: string;
      FSendingApplication: string;
      FSendingFacility: string;
      FReceivingApplication: string;
      FReceivingFacility: string;
      FMessageType: string;
      FEventType: string;
      FMessageControlID: string;
      FProcessingCode: string;
      FVersionID: string;

      FDescription: string;
      FOBRSegment: string;
      }

      DataNode := DeviceNode.ChildNodes.FindNode(FIELDSEP);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
      begin
        Pipe := DataNode.NodeValue;
        if Length(Pipe) > 0 then
          Device.FFieldSeperator := Pipe[1];
      end;

      DataNode := DeviceNode.ChildNodes.FindNode(ENCODINGCHARS);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FEncodingChars := DataNode.NodeValue;

      DataNode := DeviceNode.ChildNodes.FindNode(SENDINGAPP);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FSendingApplication := DataNode.NodeValue;

      DataNode := DeviceNode.ChildNodes.FindNode(SENDINGFACILITY);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FSendingFacility := DataNode.NodeValue;

      DataNode := DeviceNode.ChildNodes.FindNode(RECEIVINGAPP);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FReceivingApplication := DataNode.NodeValue;

      DataNode := DeviceNode.ChildNodes.FindNode(RECEIVINGFACILITY);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FReceivingFacility := DataNode.NodeValue;

      DataNode := DeviceNode.ChildNodes.FindNode(MESSAGETYPE);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FMessageType := DataNode.NodeValue;

      DataNode := DeviceNode.ChildNodes.FindNode(EVENTTYPE);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FEventType := DataNode.NodeValue;

      DataNode := DeviceNode.ChildNodes.FindNode(MESSAGECONTROL_ID);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FMessageControlID := DataNode.NodeValue;

      DataNode := DeviceNode.ChildNodes.FindNode(PROCESSING_CODE);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FProcessingCode := DataNode.NodeValue;

      DataNode := DeviceNode.ChildNodes.FindNode(SENDALLATONCE);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FSendAllObservationsAtOnce := DataNode.NodeValue;

      DataNode := DeviceNode.ChildNodes.FindNode(FREQUENCY);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FInterval := DataNode.NodeValue;

      DataNode := DeviceNode.ChildNodes.FindNode(VERSION_ID);
      if assigned(DataNode) and (DataNode.NodeValue <> null) then
        Device.FVersionID := DataNode.NodeValue;

      DTypesNode := DeviceNode.ChildNodes.FindNode(DTYPES);
      if assigned(DTypesNode) then
      begin
        for J := 0 to DTypesNode.ChildNodes.Count - 1 do
        begin
          Observation := Device.FObservationList.Add;
          DTypeNode := DTypesNode.ChildNodes[J];
          {
          FTYPE = 'type';
          DESCRIPTION = 'description';
          PRECISION = 'precision';
          ERRORDATA = 'errordata';
          ERRORPERCENTAGE = 'errorpercentage';
          VALUETYPE = 'valuetype';
          RANGEHI = 'rangehi';
          RANGELO = 'rangelo';
          OBSERVATIONIDENTIFIER = 'observationidentifier';
          OBSERVATIONSUBID = 'observationsubid';
          OBSERVATIONVALUE = 'observationvalue';
          ABNORMALFLAGS = 'abnormalflags';
          PROBABILITY = 'probability';
          ABNORMALNATURE = 'abnormalnature';
          OBSERVATIONRESULTSTATUS 'observationresultstatus';
          REFRANGEEFFDATE = 'refrangeeffdate';
          USERDEFINEDACCESSCHECKS = 'userdefinedaccesschecks';
          PRODUCERID = 'producerid';
          RESPONSIBLEOBSERVER = 'responsibleobserver';
          OBSERVATIONMETHOD = 'observationmethod';
          EQUIPMENTINSTANCEIDENTIFIER = 'equipmentinstanceidentifier';
          }
          DataNode := DTypeNode.ChildNodes.FindNode(FTYPE);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FType := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(DESCRIPTION);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FDescription := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(PRECISION);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FPrecision := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(ERRORDATA);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FErrorData := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(ERRORPERCENTAGE);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FErrorPercentage := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(RANGEHI);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FRangeHi := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(RANGELO);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FRangeLo := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(OBSERVATIONIDENTIFIER);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FObservationIdentifier := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(OBSERVATIONSUBID);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FObservationSubID := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(OBSERVATIONVALUE);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FObservationValue := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(ABNORMALFLAGS);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FAbnormalFlags := StringToAbnormalTypes(DataNode.NodeValue);

          DataNode := DTypeNode.ChildNodes.FindNode(PROBABILITY);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FProbability := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(FREQUENCY);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FFrequency := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(ERRORPERCENTAGE);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FErrorPercentage := DataNode.NodeValue;

          DataNode := DTypeNode.ChildNodes.FindNode(DIRECTVALUE);
          if assigned(DataNode) and (DataNode.NodeValue <> null) then
            Observation.FDirectValue := DataNode.NodeValue;
        end;
      end;
    end;
  finally
    XMLDoc := nil;
  end;
end;

procedure TDeviceList.Save(AFilename: string);
var
  XMLDoc: IXMLDocument;
  RootNode: IXMLNode;
  I,J: Integer;
  DeviceNode: IXMLNode;
  DTypesNode: IXMLNode;
  DTypeNode: IXMLNode;
  ChildNode: IXMLNode;
begin
  XMLDoc := TXMLDocument.Create(nil);
  try
    DefaultDOMVendor := 'MSXML';

    XMLDoc.Options := XMLDoc.Options + [doNodeAutoIndent];
    XMLDoc.Active := true;
    XMLDoc.Encoding := 'UTF-8';

    RootNode := XMLDoc.CreateElement(DEVICES, '');
    XMLDoc.DocumentElement := RootNode;

    for I := 0 to Count - 1 do
    begin
      DeviceNode := RootNode.AddChild(Device);

      ChildNode := DeviceNode.AddChild(NAME);
      ChildNode.NodeValue := Items[I].FDescription;

      ChildNode := DeviceNode.AddChild(OBRSEGMENT);
      ChildNode.NodeValue := Items[I].FOBRSegment;

      ChildNode := DeviceNode.AddChild(ORCSEGMENT);
      ChildNode.NodeValue := Items[I].FORCSegment;

      ChildNode := DeviceNode.AddChild(FIELDSEP);
      ChildNode.NodeValue := Items[I].FFieldSeperator;

      ChildNode := DeviceNode.AddChild(ENCODINGCHARS);
      ChildNode.NodeValue := Items[I].FEncodingChars;

      ChildNode := DeviceNode.AddChild(SENDINGAPP);
      ChildNode.NodeValue := Items[I].FSendingApplication;

      ChildNode := DeviceNode.AddChild(SENDINGFACILITY);
      ChildNode.NodeValue := Items[I].FSendingFacility;

      ChildNode := DeviceNode.AddChild(RECEIVINGAPP);
      ChildNode.NodeValue := Items[I].FReceivingApplication;

      ChildNode := DeviceNode.AddChild(RECEIVINGFACILITY);
      ChildNode.NodeValue := Items[I].FReceivingFacility;

      ChildNode := DeviceNode.AddChild(MESSAGETYPE);
      ChildNode.NodeValue := Items[I].FMessageType;

      ChildNode := DeviceNode.AddChild(EVENTTYPE);
      ChildNode.NodeValue := Items[I].FEventType;

      ChildNode := DeviceNode.AddChild(MESSAGECONTROL_ID);
      ChildNode.NodeValue := Items[I].FMessageControlID;

      ChildNode := DeviceNode.AddChild(PROCESSING_CODE);
      ChildNode.NodeValue := Items[I].FProcessingCode;

      ChildNode := DeviceNode.AddChild(VERSION_ID);
      ChildNode.NodeValue := Items[I].FVersionID;

      ChildNode := DeviceNode.AddChild(SENDALLATONCE);
      ChildNode.NodeValue := Items[I].FSendAllObservationsAtOnce;

      ChildNode := DeviceNode.AddChild(FREQUENCY);
      ChildNode.NodeValue := Items[I].FInterval;

      DTypesNode := DeviceNode.AddChild(DTYPES);
      for J := 0 to Items[I].FObservationList.Count - 1 do
      begin
        DTypeNode := DTypesNode.AddChild(DTYPE);
        ChildNode := DTypeNode.AddChild(FTYPE);
        ChildNode.NodeValue := Items[I].FObservationList[J].FType;
        ChildNode := DTypeNode.AddChild(DESCRIPTION);
        ChildNode.NodeValue := Items[I].FObservationList[J].FDescription;
        ChildNode := DTypeNode.AddChild(PRECISION);
        ChildNode.NodeValue := Items[I].FObservationList[J].FPrecision;
        ChildNode := DTypeNode.AddChild(ERRORDATA);
        ChildNode.NodeValue := Items[I].FObservationList[J].FErrorData;
        ChildNode := DTypeNode.AddChild(ERRORPERCENTAGE);
        ChildNode.NodeValue := Items[I].FObservationList[J].FErrorPercentage;
        ChildNode := DTypeNode.AddChild(RANGEHI);
        ChildNode.NodeValue := Items[I].FObservationList[J].FRangeHi;
        ChildNode := DTypeNode.AddChild(RANGELO);
        ChildNode.NodeValue := Items[I].FObservationList[J].FRangeLo;
        ChildNode := DTypeNode.AddChild(OBSERVATIONIDENTIFIER);
        ChildNode.NodeValue := Items[I].FObservationList[J].FObservationIdentifier;
        ChildNode := DTypeNode.AddChild(OBSERVATIONSUBID);
        ChildNode.NodeValue := Items[I].FObservationList[J].FObservationSubID;
        ChildNode := DTypeNode.AddChild(OBSERVATIONVALUE);
        ChildNode.NodeValue := Items[I].FObservationList[J].FObservationValue;
        ChildNode := DTypeNode.AddChild(ABNORMALFLAGS);
        ChildNode.NodeValue := AbnormalTypesToString(Items[I].FObservationList[J].FAbnormalFlags);
        ChildNode := DTypeNode.AddChild(FREQUENCY);
        ChildNode.NodeValue := Items[I].FObservationList[J].FFrequency;
        ChildNode := DTypeNode.AddChild(PROBABILITY);
        ChildNode.NodeValue := Items[I].FObservationList[J].FProbability;
        ChildNode := DTypeNode.AddChild(ABNORMALNATURE);
        ChildNode.NodeValue := Items[I].FObservationList[J].FAbnormalNature;
        ChildNode := DTypeNode.AddChild(OBSERVATIONRESULTSTATUS);
        ChildNode.NodeValue := Items[I].FObservationList[J].FObservationResultStatus;
        ChildNode := DTypeNode.AddChild(REFRANGEEFFDATE);
        ChildNode.NodeValue := Items[I].FObservationList[J].FRefRangeEffDate;
        ChildNode := DTypeNode.AddChild(USERDEFINEDACCESSCHECKS);
        ChildNode.NodeValue := Items[I].FObservationList[J].FUserDefinedAccessChecks;
        ChildNode := DTypeNode.AddChild(PRODUCERID);
        ChildNode.NodeValue := Items[I].FObservationList[J].FProducerID;
        ChildNode := DTypeNode.AddChild(RESPONSIBLEOBSERVER);
        ChildNode.NodeValue := Items[I].FObservationList[J].FResponsibleObserver;
        ChildNode := DTypeNode.AddChild(OBSERVATIONMETHOD);
        ChildNode.NodeValue := Items[I].FObservationList[J].FObservationMethod;
        ChildNode := DTypeNode.AddChild(EQUIPMENTINSTANCEIDENTIFIER);
        ChildNode.NodeValue := Items[I].FObservationList[J].FEquipmentInstanceIdentifier;
        ChildNode := DTypeNode.AddChild(DIRECTVALUE);
        ChildNode.NodeValue := Items[I].FOBservationList[J].FDirectValue;
      end;
    end;
    XMLDoc.SaveToFile(AFilename);
  finally
    XMLDoc := nil;
  end;
end;

procedure TDeviceList.SetDelimiters(const Value: string);
var
  I : Integer;
begin
  if FDelimiters <> Value then
  begin
    for I := 0 to Count -1 do
      Items[I].FObservationList.FDelimiters := Value;
    FDelimiters := Value;
  end;
end;

procedure TDeviceList.SetItem(Index: Integer; Value: TDevice);
begin
  inherited SetItem(Index, Value);
end;

{ TDevice }

constructor TDevice.Create(Collection: TCollection);
begin
  inherited;
  FObservationList := TObservationList.Create(TObservation);
  FObservationList.Parent := Self;

end;

destructor TDevice.Destroy;
begin
  FreeAndNil(FObservationList);
  inherited;
end;


{ TDeviceType }

constructor TObservation.Create(Collection: TCollection);
begin
  inherited;
end;

destructor TObservation.Destroy;
begin
  inherited;
end;

function TObservation.GetSegment: string;
var
  Delim: char;
  //RepSeg: char;
  function GetRange(RangeLo, RangeHi: single): string;
  begin
    if RangeLo  < 0 then
    begin
      result := '<' + FloatToStr(RangeHi);
      exit;
    end;
    if RangeHi < 0  then
    begin
      result := '>' + FloatToStr(RangeLo);
      exit;
    end;
    result := FloatToStr(RangeLo) + '-' + FloatToStr(RangeHi);
  end;
begin
  {
    Returns an OBX HL7 Segment.  See HL7 v 2.5 Spec chapter 7.4.2 for more
    info.  We're also going to add variables here for all of the defined fields
    in the HL7 Segment whether we use them or not.  After all, we may need them
    later, and it will be a lot easier to work with them later if we add them
    now.
  }

  if FDirectValue <> '' then
  begin
    result := Replace(FDirectValue);
  end
  else
  begin

    // See if this needs to be an error or not.
    if RandomRange(1, 100) < FErrorPercentage then
    begin
      //if Trim(FErrorData) <> '' then
    end
    else
    begin

    end;

    Delim := TObservationList(Collection).FParent.FFieldSeperator;
    //RepSeg := TObservationList(Collection).FParent.FEncodingChars[1];

    result :=
      OBX + Delim +
      VALUE_TYPE_ID[ord(FType)] + Delim +
      FObservationIdentifier + Delim +
      FObservationSubID + Delim +
      FObservationValue + Delim +
      // Unit of measure +
      GetRange(FRangeLo, FRangeHi) + Delim +
      AbnormalTypesToString(FAbnormalFlags) + Delim +
      FProbability + Delim +
      FAbnormalNature + Delim +
      FObservationResultStatus + Delim +
      '' + Delim + //Date of reference Range
      FUserDefinedAccessChecks + Delim +
      Replace(DATE_TIME) + Delim +
      FProducerID + Delim +
      FResponsibleObserver + Delim +
      FObservationMethod + Delim +
      FEquipmentInstanceIdentifier + Delim +
      Replace(DATE_TIME) + Delim;
  end;
end;

function TDevice.GetHeader: string;
begin
  { TODO 1 -oblj -cHL7 Message Construction :
  We need to add persistable properties for all of the header stuff
  (like the source and target app, event type, and such like), and then
  put them together in a valid string here. }

  result :=
    MSH + FFieldSeperator +
    FEncodingChars + FFieldSeperator +
    FSendingApplication + FFieldSeperator +
    FSendingFacility + FFieldSeperator +
    FReceivingApplication + FFieldSeperator +
    FReceivingFacility + FFieldSeperator +
    Replace(REP_TODAY) + FFieldSeperator + FFieldSeperator +
    FMessageType + FEncodingChars[1] + FEventType + FFieldSeperator +
    FMessageControlID + FFieldSeperator +
    FProcessingCode + FFieldSeperator +
    FVersionID + FFieldSeperator;
end;

{ TObservationList }

function TObservationList.Add: TObservation;
begin
  result := TObservation(inherited Add);
end;

destructor TObservationList.Destroy;
begin
  inherited;
end;

function TObservationList.GetItem(Index: Integer): TObservation;
begin
  result := TObservation(inherited GetItem(Index));
end;

function TObservationList.Insert(Index: Integer): TObservation;
begin
  result := TObservation(inherited Insert(Index));
end;

procedure TObservationList.SetItem(Index: Integer; Value: TObservation);
begin
  inherited SetItem(Index, Value);
end;

{ TUnownedDeviceList }

function TUnownedDeviceList.Add(Device: TDevice): Integer;
begin
  result := inherited Add(Device);
end;

function TUnownedDeviceList.Extract(Device: TDevice): TDevice;
begin
  result := TDevice(inherited Extract(Device));
end;

function TUnownedDeviceList.First: TDevice;
begin
  result := TDevice(inherited First);
end;

function TUnownedDeviceList.Get(Index: Integer): TDevice;
begin
  result := TDevice(inherited Get(Index));
end;

function TUnownedDeviceList.IndexOf(Device: TDevice): Integer;
begin
  result := inherited IndexOf(Device);
end;

procedure TUnownedDeviceList.Insert(Index: Integer; Device: TDevice);
begin
  inherited Insert(Index, Device);
end;

function TUnownedDeviceList.Last: TDevice;
begin
  result := TDevice(inherited Last);
end;

procedure TUnownedDeviceList.Put(Index: Integer; Device: TDevice);
begin
  inherited Put(Index, Device);
end;

function TUnownedDeviceList.Remove(Device: TDevice): Integer;
begin
  result := inherited Remove(Device);
end;

end.
